@@ -1,30 +1,15 @@
 ;;; -*- Mode: Emacs-Lisp -*-
 
 ;;; ilisp-snd.el --
-
-;;; This file is part of ILISP.
-;;; Version: 5.10.1
-;;;
-;;; Copyright (C) 1990, 1991, 1992, 1993 Chris McConnell
-;;;               1993, 1994 Ivan Vasquez
-;;;               1994, 1995, 1996 Marco Antoniotti and Rick Busdiecker
-;;;               1996-2000 Marco Antoniotti and Rick Campbell
-;;;
-;;; Other authors' names for which this Copyright notice also holds
-;;; may appear later in this file.
-;;;
-;;; Send mail to 'majordomo@cons.org' to be included in the
-;;; ILISP mailing list. 'ilisp@cons.org' is the general ILISP
-;;; mailing list were bugs and improvements are discussed.
-;;;
-;;; ILISP is freely redistributable under the terms found in the file
-;;; COPYING.
-
-
-
-;;;
 ;;; ILISP send and support.
 ;;;
+;;; This file is part of ILISP.
+;;; Please refer to the file COPYING for copyrights and licensing
+;;; information.
+;;; Please refer to the file ACKNOWLEGDEMENTS for an (incomplete) list
+;;; of present and past contributors.
+;;;
+;;; $Id: ilisp-snd.el,v 1.6 2009/08/13 16:47:39 aidan Exp $
 
 
 ;;;%% Package / Symbol support
@@ -34,6 +19,12 @@
 ;;;
 ;;; 19990824 Marco Antoniotti
 
+;; Have compiled 21.4 code also work on XEmacs binaries with real support
+;; for multiple values, by avoiding runtime calls to #'values: 
+(eval-when-compile (when (eq 'list (symbol-function 'values))
+                     (define-compiler-macro values (&rest args)
+                       `(list ,@args))))
+
 (defvar *ILISP-default-package* "COMMON-LISP-USER")
 
 (defun ilisp-add-set-package-hook ()	; Was: add-set-package-hook
@@ -47,33 +38,47 @@
 			(ilisp-value 'ilisp-status) " :ready")
 
 		   ;; and check the Package when in ILISP
-		   (lisp-buffer-package)
-		 ))))
+		   (lisp-buffer-package)))))
 
 ;;; ilisp-check-package-advanced --
 ;;; treat DEFPACKAGE before IN-PACKAGE.
-;;;
-;;; Was: check-package-advanced
 
 (defun ilisp-check-package-advanced (hash-defpackage-forms-list 
 				     hash-in-package-forms-list)
   "Advanced check for packages in buffer.
-It handles the special case of read-time conditionals - i.e. hash plus
+It hanldes the special case of read-time conditionals - i.e. hash plus
 or minus forms - as well as normal IN-PACKAGE or DEFPACKAGE forms."
-  (let ((string
-	 (apply #'concat (nconc hash-defpackage-forms-list
-				hash-in-package-forms-list))))
-    
-    (setq string
+  (let* ((string
+          (apply #'concat (nconc hash-defpackage-forms-list
+                                 hash-in-package-forms-list)))
+         (string
 	  (format (ilisp-value 'ilisp-package-command)
 		  (format (ilisp-value 'ilisp-block-command) string)))
-    ;; (message "ilisp-check-package-advanced: '%s'.\n" string)
-    (let ((package
-	   (ilisp-send
-	    string "Finding Buffer package with hash-forms" 'pkg nil)))
-      package)))
-
+         (package
+             ;;; (string &optional message status and-go handler)
+           (let ((the-package
+                   (ilisp-send
+                    string "Finding Buffer package with hash-forms" 'pkg nil t)))
+             ;; there may have been an error: fix it.
+             (if (string-match (ilisp-value 'ilisp-error-regexp t) the-package)
+               (progn
+                 (comint-send (ilisp-process) (ilisp-value 'comint-fix-error) t t 'fix "Fixing erroneous package-situation in inferior-lisp." t nil)
+                 "nil")
+               (if the-package
+                 the-package
+                 "nil"))))
+         (case-fold-search t)
+         (npic-regexp (ilisp-value 'ilisp-no-package-in-core-regexp t)))
+    (if (and npic-regexp (string-match npic-regexp package))
+      (progn
+        (message (format "Buffer package not found. Using fallback-package: %s"
+                         (ilisp-value 'ilisp-fallback-package)))
+        (values (ilisp-value 'ilisp-fallback-package) t))
+      (progn
+        (message "Buffer package: %s" package)
+        (values package nil)))))
 
+;;;
 (defun lisp-find-hash-form ()		; Was: find-hash-form.
   "Tries to find either an hash-form, e.g. #{+|-}-form, or a regular
  in-package or defpackage form."
@@ -117,12 +122,9 @@ or minus forms - as well as normal IN-PACKAGE or DEFPACKAGE forms."
 ;;; for Scheme yet.  After all the module system of Scheme is a system
 ;;; of very high  entropy.
 ;;;
-;;; The optional 'dummy' variable is there only for backward
-;;; compatibility.
-;;;
 ;;; 19990824 Marco Antoniotti
 
-(defun lisp-buffer-package-internal (&optional dummy)
+(defun lisp-buffer-package-internal ()
   "Returns the package of the buffer.
 If SEARCH-FROM-START is T then will search from the beginning of the
 buffer, otherwise will search backwards from current point.  This
@@ -133,65 +135,63 @@ Common Lisp."
   (setq mode-line-process 'ilisp-status)
   (let* ((lisp-buffer-package t)
 	 (case-fold-search t)
-	 (hash-form-regexp (ilisp-value 'ilisp-hash-form-regexp))
+	 (hash-form-regexp (ilisp-value 'ilisp-hash-form-regexp t))
 	 (in-package-regexp (ilisp-value 'ilisp-in-package-command-string t))
 	 (defpackage-regexp (ilisp-value 'ilisp-defpackage-command-string t))
 	 (hash-in-package-forms-list nil)
 	 (hash-defpackage-forms-list nil)
          (in-package-found-p nil)
-	 (package nil))
-    (save-excursion
-      (goto-char (point-min))
+	 (package nil)
+         (should-not-cache-p nil))
+    (if (not hash-form-regexp)
+        (values nil nil)
+      (save-excursion
+        (goto-char (point-min))
 
-      ;; This should be rewritten using LOOP.
-      ;;
-      ;; 19990824 Marco Antoniotti
-      (while
-	  (let* ((hash-expr
-		  (ignore-errors (lisp-find-hash-form)))
-		 (sub-expr
-		  (and hash-expr
-		       (string-match hash-form-regexp hash-expr)
-		       (substring hash-expr (match-beginning 0)))))
-            
-            (when hash-expr
-	      (when (and sub-expr (string-match in-package-regexp sub-expr))
-                (setq in-package-found-p t)
-		(push hash-expr hash-in-package-forms-list))
-	      (when (and sub-expr (string-match defpackage-regexp sub-expr))
-		(push hash-expr hash-defpackage-forms-list))
-              (if (and in-package-found-p
-                       (null hash-defpackage-forms-list))
-                  ;; if we found one in-package before all defpackage's then we
-                  ;; are done this takes care of some bug cases.
-                  nil
-                  t))))
+        (while
+            (let* ((hash-expr
+                    (ignore-errors (lisp-find-hash-form)))
+                   (sub-expr
+                    (and hash-expr
+                         (string-match hash-form-regexp hash-expr)
+                         (substring hash-expr (match-beginning 0)))))
+              (when hash-expr
+                (cond ((string-match "(in-package\\s-*)" hash-expr)
+                       (setq should-not-cache-p t)
+                       nil)
+                      (t
+                       (when (and sub-expr (string-match in-package-regexp sub-expr))
+                         (setq in-package-found-p t)
+                         (push hash-expr hash-in-package-forms-list))
+                       (when (and sub-expr (string-match defpackage-regexp sub-expr))
+                         (push hash-expr hash-defpackage-forms-list))
+                        t)))))
 
-      (setq package
-	    (ilisp-check-package-advanced
-	     (nreverse hash-defpackage-forms-list) 
-	     (nreverse hash-in-package-forms-list)))
-      
-      (when (ilisp-value 'comint-errorp t)
-	(lisp-display-output package)
-	(error "No package")
-	nil)
-
-      (if (and package
-	       ;; There was a bug here, used to have the second *
-	       ;; outside of the parens.
-	       ;; CMUCL needs just that WITHIN the double-quotes
-	       ;; the old regexp is (string-match "[ \n\t:\"]*\\([^
-	       ;; \n\t\"]*\\)" package))
-	       (string-match "\\([\"].[^\"]*[\"]\\)" package))
+        (multiple-value-bind (package package-not-in-core-p)
+            (ilisp-check-package-advanced
+             (nreverse hash-defpackage-forms-list) 
+             (nreverse hash-in-package-forms-list))
+          (let ((should-not-cache-p (or should-not-cache-p package-not-in-core-p)))
+          ;;; RED? (when (ilisp-value 'comint-errorp t)
+          ;;;  (lisp-display-output package)
+          ;;;  (error "No package"))
+          
+            (when (and package
+                       ;; There was a bug here, used to have the second *
+                       ;; outside of the parens.
+                       ;; CMUCL needs just that WITHIN the double-quotes
+                       ;; the old regexp is (string-match "[ \n\t:\"]*\\([^
+                       ;; \n\t\"]*\\)" package))
+                       (string-match "\\([\"].[^\"]*[\"]\\)" package))
 	  
-	  (setq package
-		(substring package
-			   (1+ (match-beginning 1)) (1- (match-end 1)))))
-      ;; => without double-quotes
+              (setq package
+                    (substring package
+                               (1+ (match-beginning 1)) (1- (match-end 1)))))
+            ;; => without double-quotes
 
-      package)))
+            (values package should-not-cache-p)))))))
 
+;;;
 (defun set-package-lisp-always ()
   "Set inferior LISP to a named package.
 The package is set whether the buffer has a package or not!"
@@ -213,77 +213,16 @@ The package is set whether the buffer has a package or not!"
 ;;; Martin Atzmueller code ends here.
 ;;; --------------------------------------------------------------------------
 
-;;; Original version without read-time conditional checks and package
-;;; hacks by Martin Atzmueller.  This piece of code is left here for
-;;; documentation purporses.
 ;;;
-;;; Since: 5.9
-;;;
-;;; 19990824 Marco Antoniotti
-
-;;; (defun lisp-buffer-package-internal (search-from-start)
-;;;   "Returns the package of the buffer.
-;;; If SEARCH-FROM-START is T then will search from the beginning of the
-;;; buffer, otherwise will search backwards from current point."
-;;;   (setq mode-line-process 'ilisp-status)
-;;;   (let* ((lisp-buffer-package t)
-;;;	  (case-fold-search t)
-;;;	  (regexp (ilisp-value 'ilisp-package-regexp t))
-;;;	  (spec
-;;;	   (and regexp
-;;;		(save-excursion
-;;;		  (when (or (and search-from-start
-;;;				 (goto-char (point-min))
-;;;				 (re-search-forward regexp nil t))
-;;;			    (re-search-backward regexp nil t))
-;;;		      (buffer-substring (match-beginning 0)
-;;;					(progn 
-;;;					  (goto-char (match-beginning 0))
-;;;					  (forward-sexp)
-;;;					  (point)))))))
-;;;	  (str (format (ilisp-value 'ilisp-package-command) spec))
-;;;	  (package (and spec
-;;;			(ilisp-send str "Finding buffer package" 'pkg)))
-;;;	  )
-;;;     ;; 19990806 Marco Antoniotti
-;;;     ;; Debugging
-;;;     ;;
-;;;     ;; (message "ILISP Package stuff: spec %s" spec)
-;;;     ;; (message "ILISP Package stuff: str %s" str)
-;;;     (cond ((ilisp-value 'comint-errorp t)
-;;;	    (lisp-display-output package)
-;;;	    (error "No package"))
-;;;	   ((and package 
-;;;		 ;; There was a bug here, used to have the second *
-;;;		 ;; outside of the parens.
-;;;		 (string-match "[ \n\t:\"]*\\([^ \n\t\"]*\\)" package))
-;;;	    (setq package
-;;;		  (substring package
-;;;			     (match-beginning 1) (match-end 1)))))
-;;;     package))
-
-
 (defun lisp-buffer-package ()
   "Return the package for this buffer.
 The package name is a string. If there is none, return NIL.  This
 caches the package unless 'ILISP-DONT-CACHE-PACKAGE' is non-nil, so
 calling this more than once is cheap."
   (cond ((and (not (eq buffer-package 'not-yet-computed))
-	      (null lisp-dont-cache-package)) 
+	      (null lisp-dont-cache-package))
 	 buffer-package)
 	(ilisp-completion-package ilisp-completion-package)
-	(lisp-dont-cache-package
-	 ;; Refind the package each time.
-	 (let ((package (lisp-buffer-package-internal nil)))
-	   (message "")
-	   (setq buffer-package 'not-yet-computed)
-	   (when package
-	     (setq mode-name
-		   (concat 
-		    (or buffer-mode-name 
-			(setq buffer-mode-name mode-name))
-		    ":" package)))
-	   package))
 	((or lisp-buffer-package 
 	     (memq major-mode ilisp-modes)
 	     (not (memq major-mode lisp-source-modes)))
@@ -291,16 +230,17 @@ calling this more than once is cheap."
 	(t
 	 (make-local-variable 'buffer-package)
 	 (make-local-variable 'buffer-mode-name)
-	 (let ((package (lisp-buffer-package-internal t)))
-	   (message "")
-	   (setq buffer-package package)
+         (multiple-value-bind (package should-not-cache-p)
+             (lisp-buffer-package-internal)
+	   (setq buffer-package (if (or should-not-cache-p lisp-dont-cache-package)
+                                    'not-yet-computed package))
 	   ;; Display package in mode line
 	   (when package 
 	     (setq mode-name
 		   (concat (or buffer-mode-name
 			       (setq buffer-mode-name mode-name))
-			   ":" buffer-package)))
-	   buffer-package))))
+			   ":" package)))
+	   package))))
 
 
 ;;;
@@ -347,23 +287,24 @@ With prefix specify manually."
     (lisp-buffer-package)))
 
 
-
 ;;;%Interface functions
 ;;;%%Symbols
 (defun lisp-string-to-symbol (string)
   "Convert STRING to a symbol, (package delimiter symbol).
 'package' is either package:symbol or from the current buffer."
-  (let* ((start (string-match ":+" string))
-	 (end (if start (match-end 0))))
+  (let* ((start (if (ilisp-value 'ilisp-package-separator-regexp t)
+		    (string-match (ilisp-value 'ilisp-package-separator-regexp t)
+                                  string)))
+         (end (if start (match-end 0))))
     (if start
-	(lisp-symbol
-	 (if (= start 0)
-	     ""
-	   (substring string 0 start))
-	 (substring string start end)
-	 (substring string end))
+        (lisp-symbol
+         (if (= start 0)
+             ""
+           (substring string 0 start))
+         (substring string start end)
+         (substring string end))
       (let ((package (lisp-buffer-package)))
-	(lisp-symbol package (if package "::") string)))))
+        (lisp-symbol package (if package "::") string)))))
 
 ;;;
 (defun lisp-symbol-to-string (symbol)
@@ -435,7 +376,6 @@ is no current sexp, return NIL."
 				 (progn (forward-sexp 1) (point)))))
 	  (error nil))))))
 
-
 ;;;
 (defun lisp-defun-name ()
   "Return the name of the current defun."
@@ -491,8 +431,6 @@ If FILE is NIL, the entry will be removed."
       (comint-send-code (get-buffer-process (current-buffer))
 			'ilisp-done-init)
     (when ilisp-initializing
-      ;; 19990806 Unknown Author (blame Marco Antoniotti for this)
-      ;; (message "Finished initializing %s" (car ilisp-dialect)))
       (unless comint-errorp
 	(message "Finished initializing %s" (car ilisp-dialect)))
       (setq ilisp-initializing nil
@@ -500,6 +438,38 @@ If FILE is NIL, the entry will be removed."
 	    (cons (buffer-name (current-buffer)) ilisp-initialized)))))
 
 ;;;
+
+(defun comint-send-code-init-function ()
+  "Stuff that is executed for initialization in the ilisp process buffer."
+  (let ((files ilisp-load-inits)
+        (done nil))
+    (unwind-protect
+      (progn
+        (when (not ilisp-init-binary-extension)
+          (setq ilisp-init-binary-extension 
+                  ilisp-binary-extension))
+
+        (dolist (file files)
+          (let ((load-file 
+                  (let ((source
+                          (expand-file-name (cdr file)
+                                            ilisp-*directory*))
+                        (binary
+                          (expand-file-name
+                           (lisp-file-extension (cdr file)
+                                                ilisp-binary-extension)
+                           ilisp-*directory*)))
+                    (if (file-newer-than-file-p binary source)
+                      binary
+                      source))))
+            (ilisp-load-or-send (file-name-hack load-file))))
+        (comint-send-code (ilisp-process)
+                          'ilisp-done-init)
+        (setq done t))                         
+      (unless done
+        (setq ilisp-initializing nil)
+        (abort-commands-lisp)))))
+
 (defun ilisp-init-internal (&optional sync)
   "Send all of the stuff necessary to initialize."
   (unwind-protect
@@ -515,41 +485,7 @@ If FILE is NIL, the entry will be removed."
 	;; This gets executed in the process buffer
 	(comint-send-code
 	 (ilisp-process)
-	 (function (lambda ()
-		     (let ((files ilisp-load-inits)
-			   (done nil))
-		       (unwind-protect
-			   (progn
-			     (when (not ilisp-init-binary-extension)
-			       (setq ilisp-init-binary-extension 
-				     ilisp-binary-extension))
-
-			     (dolist (file files)
-                               (let ((load-file 
-                                      (let ((source
-                                             (expand-file-name (cdr file)
-                                                               ilisp-*directory*))
-                                            (binary
-                                             (expand-file-name
-                                              (lisp-file-extension (cdr file)
-                                                                   ilisp-binary-extension)
-                                              ilisp-*directory*)))
-                                        (if (file-newer-than-file-p binary source)
-                                            binary
-                                          source))))
-                                 (ilisp-load-or-send
-                                  load-file)))
-                             
-                             (comint-send-code (ilisp-process)
-                                               'ilisp-done-init)
-                             (setq done t))
-                         
-                         (unless done
-                           (setq ilisp-initializing nil)
-                           (abort-commands-lisp))
-                         
-                         )))))
-        
+         (function comint-send-code-init-function))
         (set-ilisp-value 'ilisp-initializing t)) ; progn
     
     (unless (ilisp-value 'ilisp-initializing t)
@@ -585,7 +521,6 @@ reinitialization.  With a prefix, get the binary extensions again."
   "Synchronize with the inferior LISP and then initialize."
   (ilisp-init nil nil t))
 
-
 ;;;
 (defun call-defun-lisp (arg)
   "Put a call of the current defun in the inferior LISP and go there.
@@ -632,7 +567,6 @@ the buffer."
       (comint-kill-input)
       (insert form))))
 
-
 ;;;
 (defun ilisp-send (string &optional message status and-go handler)
   "Send STRING to the ILISP buffer.
@@ -675,8 +609,6 @@ it will be handled by HANDLER."
 		;; Martin Atzmueller 2000-01-22
 		;; this was necessary to have it work in Emacs 20.3 smoothly
                 ;; old one: t nil 'restore "Restore" t t
-		;; previous *experimental* still is:
-		;; t t 'restore "Restore" t t))
 		;; mew experimental:
 		;; t (unless dispatch 'wait) 'restore "Restore" t t))
 		t (unless dispatch 'wait) 'restore "Restore" t t))	     
